;;;; vc -- Universal Version Control interface

(define-module vc
  (use srfi-2)  ;; and-let*
  (use srfi-11) ;; let-values
  (use srfi-13) ;; string library
  (use rfc.uri)
  (use file.util)
  (use gauche.parameter)
  (extend vc.vc-base)
  (export-all))
(select-module vc)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; autoloads
(autoload vc.vc-cvs uri->vc-cvs-object path->vc-cvs-object)
(autoload vc.vc-darcs uri->vc-darcs-object path->vc-darcs-object)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; searching and dispatching

(define (sys-directory file)
  (if (file-is-directory? file)
    file
    (sys-dirname file)))

(define (vc-cvs-registered? file)
  (let ((dir (sys-directory file)))
    (if (or (file-is-directory? (build-path dir "CVS"))
            (file-is-directory? (build-path dir "CVSROOT")))
      (path->vc-cvs-object file)
      #f)))

(define (vc-darcs-registered? file)
  (let loop ((dir (sys-directory file)))
    (if (file-is-directory? (build-path dir "_darcs"))
      (path->vc-darcs-object file)
      (if (string=? dir "/")
        #f
        (loop (sys-dirname dir))))))

;; list of tests to find if a file is registered in a given vcs
(define vc-backend-tests
  (make-parameter (list vc-cvs-registered?
                        vc-darcs-registered?)))

(define (string->module-name str)
  (uri-encode-string str))

;; allow for dynamic addition and management of new vc methods
(define (uri->vc-repository uri)
  (let ((obj (uri->vc-object uri)))
    (cond
      ((is-a? obj <vc-repository>)
       obj)
      ((is-a? obj <vc-module>)
       (and (slot-bound? obj 'repository)
            (slot-ref obj 'repository)))
      ((is-a? obj <vc-data>)
       (and (slot-bound? obj 'module)
            (and-let* ((vc-module (slot-ref obj 'module)))
              (and (is-a? vc-module <vc-module>)
                   (slot-bound? vc-module 'repository)
                   (slot-ref vc-module 'repository)))))
      (else
       (error "invalid repository:" obj)))))

(define (path->vc-object file)
  (let loop ((tests (vc-backend-tests)))
    (if (null? tests)
      #f
      (or ((car tests) file)
          (loop (cdr tests))))))

(define (uri->vc-object uri)
  (let-values (((scheme specific) (uri-scheme&specific uri)))
    (if scheme
      (case (string->symbol (string-downcase scheme))
        ((cvs)   (uri->vc-cvs-object uri))
        ((darcs) (uri->vc-darcs-object uri))
        (else (error "unsupported vcs:" scheme)))
      (path->vc-object specific))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; basic operations

;; import a new module.  currently path must be an existing local path.

(define-method vc-new-module ((repos <vc-repository>) path)
  (error "new-module <vc-repository> is abstract"))

(define-method vc-new-module ((repos <string>) path . keys)
  (apply vc-new-module (uri->vc-repository repos) path keys))

;;  checkout module directory

(define-method vc-checkout ((repos <string>) path . keys)
  (apply vc-checkout (uri->vc-repository repos) path keys))

;;  new-file/dir in module

(define-method vc-add ((path <string>) . keys)
  (apply vc-add (uri->vc-object path) keys))

;;  update (file/dir in) module (optionally handle merges)

(define-method vc-update ((path <string>) . keys)
  (apply vc-update (uri->vc-object path) keys))

;;  edit (file/dir in) module (noop by default)

(define-method vc-edit ((path <string>) . keys)
  (apply vc-edit (uri->vc-object path) keys))

(define-method vc-edit ((path <vc-data>) . keys)
  #f)

;;  commit (file/dir in) module (optionally confirm up-to-date)

(define-method vc-commit ((path <string>) . keys)
  (apply vc-commit (uri->vc-object path) keys))

;;  remove file/dir in module

(define-method vc-remove ((path <string>) . keys)
  (apply vc-remove (uri->vc-object path) keys))

;;  revert file/dir in module (can usually be implemented as local remove+update)

(define-method vc-revert ((path <string>) . keys)
  (apply vc-revert (uri->vc-object path) keys))

(define-method vc-revert ((obj <vc-data>) . keys)
  (let ((path (slot-ref obj 'path)))
    (if (sys-remove path)
      (apply vc-update obj keys)
      (error "couldn't remove local copy of: " path))))

;;  rename file/dir in module (can usually be implemented as remove+add)

(define-method vc-rename ((source <string>) (dest <string>) . keys)
  (apply vc-rename (uri->vc-object source) (uri->vc-object dest) keys))

(define-method vc-rename ((source <vc-data>) (dest <vc-data>) . keys)
  (error "rename not yet supported"))

;;  tag (file/dir) (optionally as a branch)

(define-method vc-tag ((path <string>) name . keys)
  (apply vc-tag (uri->vc-object path) name keys))

;;  log of (file/dir in) module (unspecified ok, gives empty log)

;; log format
;;
;;   ((file-info (rev-info message) ...) ...)
;;
;; where the info is an alist

(define-method vc-log ((path <string>) . keys)
  (apply vc-log (uri->vc-object path) keys))

(define-method vc-log ((path <vc-data>) . keys)
  '())

(define-method vc-versions ((path <string>) . keys)
  (apply vc-versions (uri->vc-object path) keys))

(define-method vc-versions ((path <vc-data>) . keys)
  '())

;;  diff (file/dir in) module

(define-method vc-diff ((path <string>) . keys)
  (apply vc-diff (uri->vc-object path) keys))

;;  history (of repository operations) (unspecified ok, gives empty history)

(define-method vc-history ((path <string>) . keys)
  (apply vc-history (uri->vc-object path) keys))

(define-method vc-history ((path <vc-data>) . keys)
  '())

;; port access and fetching

(define-method vc-open-input ((path <string>))
  (vc-open-input (uri->vc-object path)))

;; fetch as string (defaults to port->string on vc-open-input)

(define-method vc-fetch ((path <string>))
  (vc-fetch (uri->vc-object path)))

(define-method vc-fetch ((obj <vc-object>))
  (and-let* ((p (vc-open-input path)))
    (port->string p)))

(provide "vc")

