;;;; vc.vc-cvs -- CVS version control backend

(define-module vc.vc-cvs
  (use srfi-2)   ;; and-let*
  (use srfi-11)  ;; let-values
  (use srfi-13)  ;; string library
  (use srfi-19)  ;; date/time library
  (use rfc.uri)
  (use file.util)
  (use gauche.process)
  (extend vc.vc-base)
  (export-all))
(select-module vc.vc-cvs)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; configuration

(define *cvs-program* "cvs")
(define *cvs-vendor-tag* (sys-uid->user-name (sys-getuid)))
(define *cvs-release-tag* "start")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; classes

(define-class <vc-cvs-object> (<vc-object>) ())
(define-class <vc-cvs-repository> (<vc-cvs-object> <vc-repository>) ())
(define-class <vc-cvs-module> (<vc-cvs-object> <vc-module>) ())
(define-class <vc-cvs-data> (<vc-cvs-object> <vc-data>) ())
(define-class <vc-cvs-dir> (<vc-cvs-data> <vc-dir>) ())
(define-class <vc-cvs-file> (<vc-cvs-data> <vc-file>) ())

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; utils

(define (shell-escape str)
  (regexp-replace-all #/([^\/\-\+\w\.])/ str "\\\\\\1"))

(define (fmt str . args)
  (apply format str (map (lambda (x) (if (string? x) (shell-escape x) x))
                         args)))

(define (vc-cvs-command name repos keys)
  (let-keywords* keys
      ((verbose? #f)
       (quiet? #t)
       (tempdir #f)
       (editor #f)
       (rcfile? #t) ;; should be #f?
       (history? #t)
       (compression #f)
       ;; require GSSAPI
       (authenticate? #f)
       (encrypt? #f)
       )
    (let* ((repository (and repos (repository-of repos)))
           (cvsroot (and repository (vc-cvs-object->cvsroot repository))))
      (string-append
       *cvs-program*
       (if (and cvsroot (not (equal? cvsroot "")))
         (fmt " -d ~A" cvsroot) "")
       (if verbose? "" " -q")
       (if quiet?   " -Q" "")
       (if tempdir (fmt " -T ~A" tempdir) "")
       (if editor (fmt " -e ~A" editor) "")
       (if rcfile? "" " -f")
       (if history? "" " -l")
       (if authenticate? " -a" "")
       (if encrypt? " -x" "")
       (if compression (fmt " -z ~D" compression) "")
       " " name))))

(define (vc-cvs-execute command . opt-path)
  (let ((path (get-optional opt-path #f)))
    (sys-system (string-append (if path (fmt "cd ~A; " path) "") command))))

(define (vc-cvs-parse command parser . opt-path)
  (let ((path (get-optional opt-path #f)))
    (call-with-input-process
        (string-append (if path (fmt "cd ~A; " path) "") command)
      parser)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; opening

(define (read-file-string file)
  (with-input-from-file file read))

;; could be a repository, module, local file or directory
(define (path->vc-cvs-object path)
  (cond
    ((file-is-directory? path)
     (if (file-exists? (build-path path "CVSROOT"))
       ;; a repository
       (make <vc-cvs-repository> :path path)
       ;; a sandbox file
       (and-let* ((repos-file (build-path path "CVS/Root"))
                  ((file-exists? repos-file))
                  (vc-repos (make <vc-cvs-repository>
                              :path (read-file-string repos-file)))
                  (module-file (build-path path "CVS/Repository"))
                  ((file-exists? module-file))
                  (vc-module (make <vc-cvs-module>
                               :repository vc-repos
                               :name (read-file-string module-file)))
                  (vc-dir (make <vc-cvs-dir>
                            :path path
                            :module vc-module)))
         vc-dir)))
    (else
     (let ((dir (sys-dirname path)))
       (if (file-exists? (build-path dir "CVS/"))
         (and-let* ((repos-file (build-path dir "CVS/Root"))
                    ((file-exists? repos-file))
                    (vc-repos (make <vc-cvs-repository>
                                :path (read-file-string repos-file)))
                    (module-file (build-path dir "CVS/Repository"))
                    ((file-exists? module-file))
                    (vc-module (make <vc-cvs-module>
                                 :repository vc-repos
                                 :name (read-file-string module-file)))
                    (vc-file (make <vc-cvs-file>
                               :path path
                               :module vc-module)))
           vc-file)
         ;; assume default CVSROOT (maybe backup and look for CVS in ..)
         (or (and-let* ((cvsroot (sys-getenv "CVSROOT"))
                        (vc-obj (uri->vc-cvs-object (string-append "cvs:" cvsroot))))
               (cond
                 ((is-a? vc-obj <vc-cvs-module>)
                  (make (if (file-is-directory? path)
                          <vc-cvs-dir> <vc-cvs-file>)
                    :path path :module vc-module))
                 ((is-a? vc-obj <vc-cvs-repository>)
                  (make (if (file-is-directory? path)
                          <vc-cvs-dir> <vc-cvs-file>)
                    :path path :module (make <vc-cvs-module> :repository vc-obj)))
                 ))
             (begin
               (warn "path->vc-cvs-object: no CVSROOT specified")
               #f)))))))

;; just consider top-level modules
(define (vc-cvs-module-name file)
  (cond
    ;; cute
    ((#/^\/?([^\/]*)/ file) => (cut <> 1))
    ;; shouldn't happen
    (else file)))

(define (uri->vc-cvs-object uri)
  (let-values (((scheme specific) (uri-scheme&specific uri)))
    (if scheme
      (rxmatch-case
          (if (equal? specific "") (or (sys-getenv "CVSROOT") "") specific)
        (#/:([^:]+):(?:([^:@]+)(?::([^:@\/]+))?@)?([^:@\/]+)(?::(\d+))?(?::([^\;\#]*))(?:\;([^\#]*))?(?:\#(.*))?$/
            (#f method user pass host port path file version)
          (let ((repos (make <vc-cvs-repository> :host host :path path)))
            (if port (slot-set! repos 'port port))
            (if user (slot-set! repos 'user user))
            (if pass (slot-set! repos 'password pass))
            (if (and file (not (equal? file "")))
              ;; should detect more acurately between files and dirs
              (let* ((vc-module (make <vc-cvs-module>
                                  :repository repos
                                  :name (vc-cvs-module-name file)))
                     (vc-file (make (if (#/\/$/ file)
                                      <vc-cvs-dir>
                                      <vc-cvs-file>)
                                :module vc-module
                                :path file)))
                (if version (slot-set! vc-file 'version version))
                vc-file)
              ;; no file, just return the repository
              repos)))
        (else
         (path->vc-cvs-object specific)))
      (path->vc-cvs-object uri))))

(define (vc-cvs-object->cvsroot obj)
  (cond
    ((is-a? obj <vc-cvs-repository>)
     (if (and (slot-bound? obj 'host) (slot-ref obj 'host))
       ;; need to distinguish between pserver, server and ext
       (string-append
        ":pserver:"
        (cond ((slot-ref obj 'user)
               =>
               (lambda (u)
                 (format
                  "~A~A@" u
                  (cond ((slot-ref obj 'password) => (cut format ":~A" <>))
                        (else "")))))
              (else ""))
        (slot-ref obj 'host)
        (cond ((slot-ref obj 'port) => (cut format ":~A" <>)) (else ""))
        ":"
        (slot-ref obj 'path))
       (slot-ref obj 'path)))
    ((and (is-a? obj <vc-cvs-module>) (slot-bound? obj 'repository))
     (vc-cvs-object->cvsroot (slot-ref obj 'repository)))
    ((and (is-a? obj <vc-data>) (slot-bound? obj 'module))
     (vc-cvs-object->cvsroot (slot-ref obj 'module)))
    ;; hmmm...
    ;;((not obj)
    ;; #f)
    ((string? obj)
     obj)
    (else
     (sys-getenv "CVSROOT"))))

(define (vc-cvs-object->uri obj)
  (string-append "cvs:" (vc-cvs-object->cvsroot obj)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic operations

;; import a new module.  currently path must be an existing local path.

(define-method vc-new-module ((repos <vc-cvs-object>) path . keys)
  (let-keywords* keys
      ((interactive? #f)
       (message (if interactive? #f
                    (format "vc-new-module: ~S" (sys-basename path))))
       (vendortag *cvs-vendor-tag*)
       (releasetag *cvs-release-tag*)
       (tag #f)
       (ignore '())
       (filter '()))
    ;; import [-options] repository vendortag releasetag...
    (let1 command
      (string-append
       (vc-cvs-command "import" repos keys)
       (if tag (fmt " -b ~A" tag) "")
       (if message (fmt " -m ~A" message) "")
       (if (pair? ignore) (string-join (map (cut fmt " -I ~A" <>) ignore)) "")
       (if (pair? filter) (string-join (map (cut fmt " -W ~A" <>) filter)) "")
       (fmt " ~A ~A ~A" (sys-basename path) vendortag releasetag))
      (vc-cvs-execute command path) )))

;;  checkout module dir

(define-method vc-checkout ((data <vc-cvs-data>) dir . keys)
  (apply vc-checkout (module-of data) path keys))

(define-method vc-checkout ((module <vc-cvs-module>) dir . keys)
  (let-keywords* keys
      ((tag #f))
    ;; checkout [options] modules...
    (let1 command
      (string-append
       (vc-cvs-command "checkout" module keys)
       (if tag (fmt " -r ~A" tag) "")
       (fmt " ~A" (name-of module)))
      (vc-cvs-execute command dir) )))

(define-method vc-checkout ((repos <vc-cvs-repository>) path . keys)
  (let-keywords* keys
      ((tag #f))
    ;; checkout [options] modules...
    (let1 command
      (string-append
       (vc-cvs-command "checkout" repos keys)
       (if tag (fmt " -r ~A" tag) "")
       (fmt " ~A" (sys-basename path)))
      (vc-cvs-execute command (sys-dirname path)) )))

;;  new-file/dir in module

(define-method vc-add ((obj <vc-cvs-data>) . keys)
  (let ((path (path-of obj)))
    (let-keywords* keys
        ((interactive? #f)
         (recursive? #t)
         (message (if interactive? #f
                      (format "vc-add: ~S" (sys-basename path)))))
      ;; add [options] files ...
      (let1 command
        (string-append
         (vc-cvs-command "add" obj keys)
         ;;(if tag (fmt " -r ~A" tag) "")
         (if message (fmt " -m ~A" message) "")
         (fmt " ~A" (sys-basename path)))
        (and (vc-cvs-execute command (sys-dirname path))
             (if (and recursive? (is-a? obj <vc-cvs-dir>))
               (for-each
                (cut apply vc-add <> keys)
                (directory-list path :children? #t :add-path? #t))
               #t))))))

;;  update (file/dir in) module (optionally handle merges)

(define-method vc-update ((obj <vc-cvs-data>) . keys)
  (let ((path (path-of obj)))
    (let-keywords* keys
        ((tag #f)
         (date #f)
         (add-directories? #f)
         (prune-directories? #f)
         (reset-tags? #f)
         (default-latest? #f)
         (overwrite-local? #f)
         (ignore '())
         (filter '())
         )
      ;; update [options] modules...
      (let1 command
        (string-append
         (vc-cvs-command "update" obj keys)
         (if tag (fmt " -r ~A" tag) "")
         (if date (fmt " -D ~A" date) "")
         (if add-directories? " -d" "")
         (if prune-directories? " -P" "")
         (if reset-tags? " -A" "")
         (if default-latest? " -f" "")
         (if overwrite-local? " -C" "")
         (if (pair? ignore) (string-join (map (cut fmt " -I ~A" <>) ignore)) "")
         (if (pair? filter) (string-join (map (cut fmt " -W ~A" <>) filter)) "")
         (fmt " ~A" (sys-basename path)))
        (vc-cvs-execute command (sys-dirname path)) ))))

;;  commit (file/dir in) module (optionally confirm up-to-date)

(define-method vc-commit ((obj <vc-cvs-data>) . keys)
  (let ((path (path-of obj)))
    (let-keywords* keys
        ((interactive? #f)
         (recursive? #t)
         (tag #f)
         (force? #f)
         (message (if interactive? #f
                      (format "vc-commit: ~S" (sys-basename path)))))
      ;; commit [options] files ...
      (let1 command
        (string-append
         (vc-cvs-command "commit" obj keys)
         (if tag (fmt " -r ~A" tag) "")
         (if recursive? "" " -l")
         (if force? " -f" "")
         (if message (fmt " -m ~A" message) "")
         (fmt " ~A" (sys-basename path)))
        (vc-cvs-execute command (sys-dirname path))))))

;;  remove file/dir in module

(define-method vc-remove ((obj <vc-cvs-data>) . keys)
  (let ((path (path-of obj)))
    ;; remove [options] files ...
    (let1 command
      (string-append
       (vc-cvs-command "remove" obj keys)
       (fmt " ~A" (sys-basename path)))
      (and (sys-remove path)
           (vc-cvs-execute command (sys-dirname path))))))

;;  tag (module/file/dir) (optionally as a branch)

(define-method vc-tag ((obj <vc-cvs-module>) name . keys)
  (let-keywords* keys
      ((tag #f)
       (date #f)
       (branch? #f)
       (default-latest? #f))
    (let1 command
      (string-append
       (vc-cvs-command "rtag" obj keys)
       (if branch? " -b" "")
       (if tag (fmt " -r ~A" tag) "")
       (if date (fmt " -D ~A" date) "")
       (if default-latest? " -f" "")
       (fmt " ~A ~A" name (name-of module)))
      (vc-cvs-execute command) )))

(define-method vc-tag ((obj <vc-cvs-data>) name . keys)
  (let ((path (path-of obj)))
    (let-keywords* keys
        ((check? #f)
         (branch? #f))
      (let1 command
        (string-append
         (vc-cvs-command "tag" obj keys)
         (if branch? " -b" "")
         (if check?  " -c" "")
         (fmt " ~A" (sys-basename path)))
        (vc-cvs-execute command (sys-dirname path)) ))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; output parsing commands

;;  log of (file/dir in) module

; (define (vc-cvs-parse-log p)
;   (let ((files '())
;         (file #f)
;         (revs '())
;         (data '())
;         (desc '())
;         (in-desc? #f))
;     (let loop ((line (read-line p)))
;       (if (eof-object? line)
;         (reverse (cons (list file (string-concatenate-reverse desc)) res))
;         (begin
;           (rxmatch-case line
;             (#/^-+$/ (#f)
;               ;; new revision
;               (cond
;                 (file
;                  (when (or (pair? data) (pair? desc))
;                    (push! revs (list data (string-concatenate-reverse desc))))
;                  (set! data '())
;                  (set! desc '()))
;                 (else ;; first rev
;                  (set! file (reverse data))
;                  (set! data '()))))
;             (#/^=+$/ (#f)
;               ;; new file
;               (when (or (pair? data) (pair? desc))
;                 (push! res (list data (string-concatenate-reverse desc)))
;                 (set! file #f)
;                 (set! data '())
;                 (set! desc '())))
;             (#/^cvs log:/ (#f)
;               ;; initial log message
;               )
;             (else (push! desc line)))
;           (loop (read-line p)) )))))

(define vc-cvs-parse-log port->string)

(define-method vc-log ((obj <vc-cvs-data>) . keys)
  (define (date-range->string d)
    (cond
      ((string? d) d)
      ((pair? d)
       (let* ((op (car d))
              (args (cdr d)))
         (if (= 1 (length args))
           (fmt " ~A~A" op (car args))
           (fmt "~A~A~A" (car args) op (cadr args)))))))
  (let ((path (path-of obj)))
    (let-keywords* keys
        ((no-messages? #f)
         (no-tags? #f)
         (name-only? #f)
         (default-branch? #f)
         (recursive? #t)
         (dates '())
         (tags '())) ;; XXX add tags
      (let1 command
        (string-append
         (vc-cvs-command "log" obj keys)
         (if no-messages? " -h" "")
         (if no-tags? " -N" "")
         (if name-only? " -R" "")
         (if default-branch? " -b" "")
         (if recursive? "" " -l")
         (if (null? dates) ""
             (string-append
              " -d " (string-join (map date-range->string dates) ";")))
         (fmt " ~A" (sys-basename path)))
        (vc-cvs-parse command vc-cvs-parse-log (sys-dirname path)) ))))

(define-method vc-versions ((obj <vc-cvs-data>) . keys)
  (let ((res '()))
    (with-input-from-string (apply vc-log obj :no-tags? #t keys)
      (lambda ()
        (port-for-each
         (lambda (line)
           (cond ((#/^revision ([\d.]+)$/ line)
                  => (lambda (m) (push! res (m 1))))))
         read-line)))
    (reverse! res)))

;;  diff (file/dir in) module

(define vc-cvs-parse-diff port->string)

(define-method vc-diff ((obj <vc-cvs-data>) . keys)
  (let ((path (path-of obj)))
    (let-keywords* keys
        ((recursive? #t))
      (let1 command
        (string-append
         (vc-cvs-command "diff" obj keys)
         (if recursive? "" " -l")
         (fmt " ~A" (sys-basename path)))
        (vc-cvs-parse command vc-cvs-parse-diff (sys-dirname path)) ))))

;;  history (of repository operations)

(define (vc-cvs-parse-history port)
  (with-input-from-port port
    (lambda ()
      (let loop ((line (read-line))
                 (res '()))
        (if (eof-object? line)
          (reverse! res)
          (rxmatch-case line
            (#/^([A-Z]) (\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}) ([+-]\d{4}) (\S+)\s+(\S+)\s+(\S+)\s+(\S+)/
                (#f type year mon day hour min tz author version file dir)
              (loop (read-line)
                    (cons (make <vc-change>
                            :file (string-append dir "/" file)
                            :author author
                            :version version
                            :time (time-second
                                   (date->time-monotonic
                                    (apply make-date 0 0
                                           (map string->number
                                                (list min hour
                                                      day mon year tz))))))
                          res)))
            (else
             (warn "invalid line in history: ~S" line)
             (loop (read-line) res))))))))

(define-method vc-history ((obj <vc-cvs-object>) . keys)
  (let-keywords* keys
      ((commits? #t)
       (everything? #f)
       (types '())
       (all-users? #f)
       (last-mod-only? #f)
       (working-dir-only? #f)
       (modules? '())
       (files '()))
    (let1 command
      (string-append
       (vc-cvs-command "history" obj keys)
       (if commits? " -c" "")
       (if everything? " -e" "")
       (if (pair? types) (string-append " -x " (string-join types "")) "")
       (if all-users? " -a" "")
       (if last-mod-only? " -l" "")
       (if working-dir-only? " -l" "")
       (if (pair? files)
         (string-append " " (string-join (map shell-escape files) " ")) ""))
      (vc-cvs-parse command vc-cvs-parse-history) )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; port access

(define-method vc-open-input ((obj <vc-cvs-file>))
  (let ((path (path-of obj))
        (version (version-of obj)))
    (let1 command
      (string-append
       (vc-cvs-command "checkout" obj '())
       " -p"
       (if version (fmt " -r ~A" version) "")
       (fmt " ~A" path))
      (open-input-process-port command) )))

(provide "vc/vc-cvs")

