;;;; vc.vc-darcs -- Darcs version control backend

(define-module vc.vc-darcs
  (use srfi-2)   ;; and-let*
  (use srfi-11)  ;; let-values
  (use srfi-13)  ;; string library
  (use srfi-19)  ;; date/time library
  (use util.list)
  ;;(use util.debug)
  (use rfc.uri)
  (use file.util)
  (use gauche.process)
  (extend vc.vc-base)
  (export-all))
(select-module vc.vc-darcs)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; configuration

(define *darcs-program* "darcs")

(define (darcs-author)
  (if (or (sys-getenv "DARCS_EMAIL")
          (sys-getenv "EMAIL"))
    #f
    (string-append (sys-uid->user-name (sys-getuid))
                   "@" (sys-gethostname))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; classes

(define-class <vc-darcs-object> (<vc-object>) ())
(define-class <vc-darcs-repository> (<vc-darcs-object> <vc-repository>) ())
(define-class <vc-darcs-module> (<vc-darcs-object> <vc-module>) ())
(define-class <vc-darcs-data> (<vc-darcs-object> <vc-data>) ())
(define-class <vc-darcs-dir> (<vc-darcs-data> <vc-dir>) ())
(define-class <vc-darcs-file> (<vc-darcs-data> <vc-file>) ())

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; utils

(define (shell-escape str)
  (regexp-replace-all #/([^\/\-\+\w\.])/ str "\\\\\\1"))

(define (fmt str . args)
  (apply format str (map (lambda (x) (if (string? x) (shell-escape x) x))
                         args)))

(define (vc-darcs-execute command . opt-path)
  (let* ((path (get-optional opt-path #f))
         (cmd (string-append "yes | (" (if path (fmt "cd ~A; " path) "") command ")")))
    ;;(format #t "execute: ~S\n" cmd)
    (sys-system cmd)))

(define (vc-darcs-parse command parser . opt-path)
  (let* ((path (get-optional opt-path #f))
         (cmd (string-append "yes | (" (if path (fmt "cd ~A; " path) "") command ")")))
    ;;(format #t "parse: ~S\n" cmd)
    (call-with-input-process cmd parser)))

(define (vc-darcs-command name keys)
  (let-keywords* keys
      ((verbose? #f))
    (format "~A ~A~A" *darcs-program* name (if verbose? " -v " ""))))

(define (vc-darcs-root+file obj)
  (and-let* ((top0 (path-of (repository-of obj)))
             (top (if (string-suffix? "/" top0) top0 (string-append top0 "/")))
             (path (path-of obj))
             (off (string-prefix-length top path)))
    (if (or (string=? top path) (string=? top0 path))
      (values top "")
      (values top (substring path off (string-length path))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; opening

(define (read-file-string file)
  (with-input-from-file file read))

(define (sys-directory x)
  (if (file-is-directory? x) x (sys-dirname x)))

(define (darcs-root path)
  (let loop ((dir (sys-normalize-pathname (sys-directory path)
                                          :absolute #t
                                          :canonicalize #t)))
    (if (file-exists? (build-path dir "_darcs"))
      dir
      (if (string=? dir "/")
        #f
        (loop (sys-dirname dir))))))

(define (path->vc-darcs-object path)
  ;;(debug "path->vc-darcs-object ~S" path)
  (and-let* ((top (darcs-root path))
             (repos (make <vc-darcs-repository> :path top))
             (mod (make <vc-darcs-module> :repository repos)))
    (if (file-is-directory? path)
      (make <vc-darcs-dir> :path path :module mod)
      (make <vc-darcs-file> :path path :module mod))))

;; darcs:<repos>;<file>#<patch>
;;   where <repos> can be a local directory or remote http server

(define (uri->vc-darcs-object uri)
  (let-values (((scheme specific) (uri-scheme&specific uri)))
    (and scheme
         (rxmatch-case specific
           (#/^.*([^\;\#]*)(?:\;([^\#]*))?(?:\#(.*))?$/
                 (#f uri file version)
                 (if file ;; file separate from repos
                   (let* ((repos (make <vc-darcs-repository> :path uri))
                          (mod (make <vc-darcs-module> :repository repos)))
                     (if (#/\/$/ file)
                       (make <vc-darcs-dir> :path path :module mod)
                       (make <vc-darcs-file> :path path :module mod)))
                   (or (path->vc-darcs-object uri)
                       (make <vc-darcs-repository> :path (sys-directory uri)))))))))

(define (vc-darcs-object->uri obj)
  (string-append "darcs:" (path-of obj)))

;; used for update, check for an optional parent repository
(define (darcs-repository obj)
  (let* ((root (path-of (repository-of obj)))
         (repos (build-path root "_darcs/prefs/repos")))
    (if (file-exists? repos)
      (with-input-from-file repos read-line)
      root)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; basic operations

;; inittree

(define-method vc-new-module ((repos <vc-darcs-object>) path . keys)
  ;;(format #t "vc-new-module ~S ~S\n" repos path)
  (let-keywords* keys
      ((auto-add? #t))
    (let1 command (vc-darcs-command "inittree" keys)
      (vc-darcs-execute command path))
    (when auto-add?
      (for-each (cut apply vc-add <> keys)
                (map (cut string-append path "/" <>)
                     (delete "_darcs" (cddr (sys-readdir path)))))
      (apply vc-commit path keys))))

;; get

(define-method vc-checkout ((obj <vc-darcs-object>) path . keys)
  (let-keywords* keys
      ((tag #f)
       (patch #f))
    (let1 command
      (string-append
       (vc-darcs-command "get" keys)
       (if tag (fmt " -t ~A" tag) "")
       (if patch (fmt " -m ~A" patch) "")
       (fmt " ~A" (path-of (repository-of obj))))
      (vc-darcs-execute command (sys-dirname path)) )))

;; add

(define-method vc-add ((obj <vc-darcs-data>) . keys)
  (receive (root file) (vc-darcs-root+file obj)
    ;;(format #t "vc-add ~S ~S ~S\n" obj root file)
    (let-keywords* keys
        ((recursive? #t))
      (let1 command
          (string-append
           (vc-darcs-command "add" keys)
           (if recursive? " -r " "")
           (fmt " ~A" (if (string-prefix? "./" file)
                        (sys-normalize-pathname file :absolute #t :canonicalize #t)
                        (if (or (string=? file ".") (string=? file ""))
                          root
                          file))))
        (vc-darcs-execute command root)))))

;; pull

(define-method vc-update ((obj <vc-darcs-data>) . keys)
  (receive (root file) (vc-darcs-root+file obj)
    (let ((repos (darcs-repository obj)))
      (let-keywords* keys
          ((patch #f)
           (all? #f)
           (ignore-times? #f))
        (let1 command
            (string-append
             (vc-darcs-command "pull" keys)
             (if patch (fmt " -m ~A" tag) "")
             (if all? " -a" "")
             (if ignore-times? " --ignore-time" "")
             (fmt " ~A" repos))
          (vc-darcs-execute command root) )))))

;; record

(define-method vc-commit ((obj <vc-darcs-data>) . keys)
  (receive (root file) (vc-darcs-root+file obj)
    ;;(format #t "vc-commit ~S ~S ~S\n" obj root file)
    (let-keywords* keys
        ((interactive? #f)
         (author (if interactive? #f (darcs-author)))
         (test? #f)
         (all? (if interactive? #f #t))
         (ask-deps? #f)
         (ignore-times? #f)
         (message (if interactive? #f (string-append "vc-commit: " file))))
      (let1 command
        (string-append
         (vc-darcs-command "record" keys)
         (if test? "" " --no-test")
         (if all? " -a" "")
         (if ask-deps? " --ask-deps" "")
         (if ignore-times? " --ignore-times" "")
         (if author (fmt " -A ~A" author) "")
         (if message (fmt " -m ~A" message) ""))
        (vc-darcs-execute command root)))))

;; remove

(define-method vc-remove ((obj <vc-darcs-data>) . keys)
  (receive (root file) (vc-darcs-root+file obj)
    (let1 command
      (string-append
       (vc-darcs-command "remove" keys)
       (fmt " ~A" file))
      (vc-darcs-execute command root))))

;; mv

(define-method vc-rename ((obj <vc-darcs-data>) dest . keys)
  (receive (root file) (vc-darcs-root+file obj)
    (let1 command
      (string-append
       (vc-darcs-command "mv" keys)
       (fmt " ~A ~A" file dest))
      (vc-darcs-execute command root))))

;; tag

(define-method vc-tag ((obj <vc-darcs-data>) name . keys)
  (receive (root file) (vc-darcs-root+file obj)
    (let-keywords* keys
        ((interactive? #f)
         (author (if interactive? #f (darcs-author)))
         (name (if interactive? #f
                   (format "vc-tag: ~S" (sys-time)))))
      (let1 command
        (string-append
         (vc-darcs-command "tag" keys)
         (if author (fmt " -A ~A" author) "")
         (if name  (fmt " -m ~A" name) ""))
        (vc-darcs-execute command root) ))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; output parsing commands

;; diff

(define vc-darcs-parse-diff port->string)

(define-method vc-diff ((obj <vc-darcs-data>) . keys)
  (receive (root file) (vc-darcs-root+file obj)
    (let-keywords* keys
        ((patch #f))
      (let1 command
        (string-append
         (vc-darcs-command "diff" obj keys)
         (if patch (fmt " -m ~A" patch) "")
         (fmt " ~A" file))
        (vc-darcs-parse command vc-darcs-parse-diff root) ))))

;; history (no darcs command, we parse the patch dir)

(define-method vc-history ((obj <vc-darcs-data>) . keys)
  (define (month-number mon)
    (case (string->symbol (string-downcase mon))
      ((jan) 1) ((feb) 2) ((mar) 3) ((apr) 4) ((may) 5) ((jun) 6)
      ((jul) 7) ((aug) 8) ((sep) 9) ((oct) 10) ((nov) 11) ((dec) 12)
      (else #f)))
  (receive (root file) (vc-darcs-root+file obj)
    (let ((dir (string-append root "/_darcs/patches")))
      (reverse
       (fold
        (lambda (f acc)
          (rxmatch-case f
            (#/^(.*)-([^-]+)-\w{3}_(\w{3})_(\d{2})_(\d{2}).(\d{2}).(\d{2})_(\w{3})_(\d{4})$/
                (#f comment author mon day hour min sec tz year)
              (cons (make <vc-change>
                      :comment comment
                      :author (regexp-replace #/9/ author "@")
                      :file (with-input-from-file
                                (string-append dir "/" f)
                              (lambda ()
                                (let loop ((line (read-line)))
                                  (cond
                                    ((eof-object? line) #f)
                                    ;; handle vt100 highlighting
                                    ((#/^(?:\e\[[\d\;]*m)?hunk(?:\e\[[\d\;]*m)? (\S+) \d+$/ line)
                                     => (cut <> 1))
                                    (else
                                     ;;(warn "no hunk: ~A" line)
                                     (loop (read-line)))))))
                      ;; XXXX need tz lookup from short name
                      :time (time-second
                             (date->time-monotonic
                              (apply make-date 0
                                     (map x->number
                                          (list sec min hour day
                                                (month-number mon)
                                                year 0))))))
                    acc))
            (else
             (unless (equal? f "pending")
               (warn "unknown patch format: ~S" f))
             acc)))
        '()
        (cddr (sys-readdir dir)))))))

(provide "vc/vc-darcs")

