
(use srfi-1)
(use srfi-2)
(use file.util)
(use gauche.test)

(test-start "CVS")

(set! *load-path* (cons ".." *load-path*))
(use vc)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; utils

;; create and return the name of a temp directory
(define (make-temp-dir)
  (and-let* ((name (sys-tmpnam)))
    (make-directory* name)
    name))

;; bind name to a newly created temp directory, execute body, then
;; remove the temp directory
(define-macro (with-temp-dir name . body)
  `(let ((,name (make-temp-dir)))
     ,@body
     (remove-directory* ,name)
     ))

;; shorthand for (with-output-to-file file (cut display contents))
(define (write-file name . opt-contents)
  (and-let* ((p (open-output-file name :if-exists :overwrite)))
    (if (pair? opt-contents)
      (display (car opt-contents) p)
      (close-output-port p))))

;; write-fs -- write a directory and file hierarchy

;; (dirname (spec ...))
;; (filename [contents])
;; (filename #t)

(define (write-fs hierarchy . opt-root)
  (define (fs-dir? x)
    (or (null? (cdr x)) (and (pair? (cdr x)) (pair? (cadr x)))))
  (let loop ((tr hierarchy)
             (rt (get-optional opt-root ".")))
    (when (pair? tr)
      (let ((name (car tr)))
        (if (fs-dir? tr)
          (let ((root2 (build-path rt name)))
            (make-directory* root2)
            (for-each (cut loop <> root2) (cdr tr)))
          (let* ((file (car tr))
                 (contents (cadr tr))
                 (writer (cond ((string? contents)
                                (lambda () (display contents)))
                               ((procedure? contents)
                                contents)
                               (else (lambda () #t))))
                 (keywords (cddr tr))
                 (if-exists (get-keyword :if-exists keywords :overwrite)))
            (with-output-to-file (build-path rt file)
              writer
              :if-exists if-exists)))))))

(define (fs-equal? a b . keywords)
  (let ((keep (get-keyword :filter keywords (lambda (x) #t))))
    (define (loop x y)
      (if (file-is-directory? x)
        (or (file-eqv? x y)
            (and (file-is-directory? y)
                 (let ((x-files (directory-list x :children? #t :add-path? #t :filter keep))
                       (y-files (directory-list y :children? #t :add-path? #t :filter keep)))
                 (and (equal? x-files y-files)
                      (every loop x-files y-files) ))))
        (file-equal? x y)))
    (loop a b)))

(define (sandbox-equal? x y)
  (fs-equal? x y :filter (lambda (f) (not (equal? (sys-basename f) "CVS")))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create a temp repository to use throughout the tests

(define repos (make-temp-dir))

(test* "init-repository" #t
  (and (sys-system (format "cvs -d '~A' init" repos)) #t))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; run tests

(test "new-module and checkout" #t
  (lambda ()
    (with-temp-dir root
      (write-fs '("hello" ("hello.c" "main (int argc, char **argv) {printf(\"hello world\\n\";)}")
                          ("README" ""))
                root)
      (vc-new-module repos (build-path root "hello"))
      (with-temp-dir sandbox
        (vc-checkout repos (build-path sandbox "hello"))
        (sandbox-equal? (build-path sandbox "hello")
                        (build-path root "hello"))))))

(test "add, commit and update" #t
  (lambda ()
    (with-temp-dir root
      (write-fs '("quine" ("quine.scm" "((lambda (x) `(,x ',x)) '(lambda (x) `(,x ',x)))")
                          ("README" ""))
                root)
      (vc-new-module repos (build-path root "quine"))
      (with-temp-dir sandbox
        (vc-checkout repos (build-path sandbox "quine"))
        (with-temp-dir sandbox2
          (vc-checkout repos (build-path sandbox2 "quine"))
          (write-fs '("quine" ("quine2.scm" "((lambda (x) `(,(reverse x) ',x)) '(`(,(reverse x) ',x) (x) lambda))"))
                    sandbox)
          (vc-add (build-path sandbox "quine/quine2.scm"))
          (vc-commit (build-path sandbox "quine/quine2.scm"))
          (vc-update (build-path sandbox2 "quine"))
          (sandbox-equal? (build-path sandbox "quine")
                          (build-path sandbox2 "quine")))))))

;; factor the above pattern
(define (test-sandboxes comment fs proc)
  (test comment #t
    (lambda ()
      (with-temp-dir root
        (write-fs fs root)
        (vc-new-module repos (build-path root (car fs)))
        (with-temp-dir sandbox1
          (vc-checkout repos (build-path sandbox1 (car fs)))
          (with-temp-dir sandbox2
            (vc-checkout repos (build-path sandbox2 (car fs)))
            (proc sandbox1 sandbox2)))))))

(test-sandboxes "modify, commit and update"
  '("c-quine" ("quine.c"  "char*f=\"char*f=%c%s%c;main(){printf(f,34,f,34,10);}%c\";main(){printf(f,34,f,34,10);}\n")
              ("README"   "make test fails")
              ("Makefile" "all: quine\n\ntest: quine\n\tif ./quine | diff -q /dev/fd/0 quine.c; then echo ok; else echo failed; fi"))
  (lambda (sandbox1 sandbox2)
    (write-fs '("c-quine" ("README" "make test works!")) sandbox1)
    (vc-commit (build-path sandbox1 "c-quine"))
    (vc-update (build-path sandbox2 "c-quine"))
    (sandbox-equal? sandbox1 sandbox2)))

(test-sandboxes "remove, commit and update"
  '("shell" ("run.sh"   "#!/bin/sh\nmake love")
            ("Makefile" ".PHONY: love"))
  (lambda (sandbox1 sandbox2)
    (vc-remove (build-path sandbox1 "shell/Makefile"))
    (vc-commit (build-path sandbox1 "shell"))
    (vc-update (build-path sandbox2 "shell"))
    (sandbox-equal? sandbox1 sandbox2)))

(test-sandboxes "modify and revert"
  '("perl" ("japh.pl" "#!/usr/bin/perl
                       $_ = q ;4a75737420616e6f74686572205065726c204861636b65720as;;
                       for (s;s;s;s;s;s;s;s;s;s;s;s)
                         {s;(..)s?;qq qprint chr 0x$1 and \\161 ssq;excess;}"))
  (lambda (sandbox1 sandbox2)
    (write-fs '("perl" ("japh.pl" "Just another Perl Hacker")) sandbox1)
    (vc-revert (build-path sandbox1 "perl/japh.pl"))
    (sandbox-equal? sandbox1 sandbox2)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; cleanup the temp repository

(remove-directory* repos)

(test-end)

